    MODULE get_numbers_mod
!!!**********************************************************************
!!!
!!!SUMMARY
!!!
!!!get_numbers is a  generic Fortran 90 routine  that prompts a user then
!!!reads   a  specified number  of  values   into  array  x.  Optionally,
!!!get_numbers checks  high and low bounds  on the numbers.   The type of
!!!each element of x and of  lo and hi must be  the same and either
!!!double precision, real, or integer.  x, lo, and hi can all be scalars
!!!or all can be arrays of rank 1.
!!!
!!!DETAILS
!!!
!!!    SUBROUTINE get_numbers(format,lo,lo_eq_ok,hi,hi_eq_ok,x,
!!!               number_wanted, failed)
!!!
!!!    CHARACTER (*), OPTIONAL, INTENT (IN) :: format
!!!
!!!If present,   a  character string  containing   a Fortran format.  The
!!!format will be printed to standard output (*)  to prompt the user.  If
!!!the argument is missing, no prompt message will be printed.
!!!
!!!      <TYPE>, OPTIONAL, INTENT (IN) :: lo(:)     or
!!!      <TYPE>, OPTIONAL, INTENT (IN) :: lo
!!!
!!!<TYPE> should be the same as the type of x: double precision, integer,
!!!or real.
!!!
!!!Low bounds for the values entered.  If lo has size 1, the single value
!!!is used to check all input values.   If lo does not  have size 1, then
!!!it should have size  of at least the  number of values to be obtained.
!!!In this case the i'th value obtained  from the user is checked against
!!!the i'th value  of lo.   If this argument  is missing,  no lower bound
!!!checking is performed.
!!!
!!!      LOGICAL, OPTIONAL, INTENT (IN) :: lo_eq_ok
!!!
!!!If .TRUE., a value can equal its low bound without triggering an error
!!!message; if .FALSE., it can't.  Default if missing is .TRUE.
!!!
!!!      <TYPE>, OPTIONAL, INTENT (IN) :: hi(:)     or
!!!      <TYPE>, OPTIONAL, INTENT (IN) :: hi
!!!
!!!High bounds for the values entered.  See lo.
!!!
!!!      LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok
!!!
!!!If   .TRUE., a value  can equal  its high bound  without triggering an
!!!error message; if .FALSE., it can't.  Default if missing is .TRUE.
!!!
!!!      <TYPE>, INTENT (OUT) :: x(:)     or
!!!      <TYPE>, INTENT (OUT) :: x
!!!
!!!<TYPE> should be one of: double precision, integer, or real. The array
!!!into which the values obtained from the user should be placed.
!!!
!!!      INTEGER, OPTIONAL, INTENT (IN) :: number_wanted
!!!
!!!The number of  values to be obtained from  the user.  If missing, this
!!!value is SIZE( x ).
!!!
!!!      LOGICAL, OPTIONAL, INTENT (OUT) :: failed
!!!
!!!If .TRUE. on exit, three attempts failed to obtain the values from the
!!!user.  If argument is missing,  this subroutine will abort (STOP) with
!!!an error message.
!!!
!!!**********************************************************************
!!! JJV
!!! JJV Debug changes:
!!! JJV
!!! JJV Changed number_wanted to use_n_wanted in READ statements
!!! JJV in array specifics
!!! JJV Added i's to write lists in bound checking in array specifics.
!!! JJV Added corresponding-element checking to the checking code in
!!! JJV array specifics.
!!! JJV Changed FORMAT statements in bound checking to reflect correct
!!! JJV error in all routines.
!!! JJV Replaced 'try again' messages before CYCLE TRY_LOOP with a
!!! JJV conditional WRITE at the top of the loop.  This also issues the
!!! JJV message only when it makes sense.
!!! JJV Removed FMTDBL from list of external routines in all routines.
!!! JJV
!!! JJV Cosmetic changes:
!!! JJV
!!! JJV Added INTENT (IN) to number_wanted declaration in comments and
!!! JJV array specifics.
!!! JJV Added IMPLICIT NONE statements to all routines.
!!! JJV Simplified the scalar routines.
!!! JJV Minor simplifications in all routines.
! .. Implicit None Statement ..
      IMPLICIT NONE
! ..
! .. Generic Interface Blocks ..
      INTERFACE get_numbers
        MODULE PROCEDURE get_double_array, get_integer_array, &
          get_real_array, get_double_scalar, get_integer_scalar, &
          get_real_scalar
      END INTERFACE
! ..
! .. Local Scalars ..
! Refer to :Parameters
      INTEGER, PRIVATE :: stdin = 5, stdout = 6
! ..
! .. Intrinsic Functions ..
      INTRINSIC KIND
! ..
! .. Parameters ..
      INTEGER, PARAMETER, PRIVATE :: dpkind = KIND(0.0D0)
! ..
! .. Private Statements ..
      PRIVATE :: prompt
! ..
    CONTAINS

!*********************************************************************

      SUBROUTINE get_double_array(format,lo,lo_eq_ok,hi,hi_eq_ok,x, &
          number_wanted,failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: number_wanted
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Array Arguments ..
        REAL (dpkind), OPTIONAL, INTENT (IN) :: hi(:), lo(:)
        REAL (dpkind), INTENT (OUT) :: x(:)
! ..
! .. Local Scalars ..
        INTEGER :: check_hi_status, check_lo_status, i, ios, i_try, &
          use_n_wanted
        LOGICAL :: use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT, SIZE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

!!!   CHECK_[LO|HI]_STATUS
!!!   -1 No check
!!!    1 Check against first and only element in bounds array
!!!   >1 Check against corresponding element in bounds array

        check_lo_status = -1
        IF (PRESENT(lo)) THEN
          check_lo_status = 1
          IF (SIZE(lo)/=1) check_lo_status = 2
        END IF

        check_hi_status = -1
        IF (PRESENT(hi)) THEN
          check_hi_status = 1
          IF (SIZE(hi)/=1) check_hi_status = 2
        END IF

        IF (PRESENT(number_wanted)) THEN
          use_n_wanted = number_wanted
        ELSE
          use_n_wanted = SIZE(x)
        END IF

TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again from the beginning.'

          IF (PRESENT(format)) WRITE (*,format)

          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) (x(i),i=1,use_n_wanted)
          IF (ios/=0) THEN
            WRITE (*,*) &
              'Some entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

CHECK_VALUES: DO i = 1, use_n_wanted

!        Check lower bound if required.

90000       FORMAT ('The ',I3, &
              '''th value entered < the lower bound allowed.'/ &
              'The value entered was: ',G20.7/'The lower bound is: ', &
              G20.7)
90100       FORMAT ('The ',I3, &
              '''th value entered <= the lower bound allowed.'/ &
              'The value entered was: ',G20.7/'The lower bound is: ', &
              G20.7)

LO_CHECK:   IF (check_lo_status>0) THEN
              IF (check_lo_status==1) THEN
                IF (use_lo_eq_ok .AND. x(i)<lo(1)) THEN
                  WRITE (*,90000) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(1)) THEN
                  WRITE (*,90100) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_lo_eq_ok .AND. x(i)<lo(i)) THEN
                  WRITE (*,90000) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(i)) THEN
                  WRITE (*,90100) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF LO_CHECK


!        Check upper bound if required.

90200       FORMAT ('The ',I3, &
              '''th value entered > the upper bound allowed.'/ &
              'The value entered was: ',G20.7/'The upper bound is: ', &
              G20.7)
90300       FORMAT ('The ',I3, &
              '''th value entered >= the lower bound allowed.'/ &
              'The value entered was: ',G20.7/'The upper bound is: ', &
              G20.7)

HI_CHECK:   IF (check_hi_status>0) THEN
              IF (check_hi_status==1) THEN
                IF (use_hi_eq_ok .AND. x(i)>hi(1)) THEN
                  WRITE (*,90200) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(1)) THEN
                  WRITE (*,90300) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_hi_eq_ok .AND. x(i)>hi(i)) THEN
                  WRITE (*,90200) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(i)) THEN
                  WRITE (*,90300) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF HI_CHECK

          END DO CHECK_VALUES

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN


        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) 'Too many failed attempts at entry in &
            &get_double_array. ABORT'
          STOP 'Too many failed attempts at entry in get_double_array. &
            &ABORT'
        END IF

      END SUBROUTINE get_double_array

!*********************************************************************

      SUBROUTINE get_double_scalar(format,lo,lo_eq_ok,hi,hi_eq_ok,x, &
          failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
! ..
! .. Scalar Arguments ..
        REAL (dpkind), OPTIONAL, INTENT (IN) :: hi, lo
        REAL (dpkind), INTENT (OUT) :: x
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Local Scalars ..
        INTEGER :: ios, i_try
        LOGICAL :: hi_present, lo_present, use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

        lo_present = PRESENT(lo)

        hi_present = PRESENT(hi)


TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again.'

          IF (PRESENT(format)) THEN
            WRITE (*,format)
          END IF
          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) x
          IF (ios/=0) THEN
            WRITE (*,*) 'Entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

!        Check lower bound if required.

90000     FORMAT ('The value entered < the lower bound allowed.'/ &
            'The value entered was: ',G20.7/'The lower bound is: ',G20.7)
90100     FORMAT ('The value entered <= the lower bound allowed.'/ &
            'The value entered was: ',G20.7/'The lower bound is: ',G20.7)

LO_CHECK: IF (lo_present) THEN
            IF (use_lo_eq_ok .AND. x<lo) THEN
              WRITE (*,90000) x, lo
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_lo_eq_ok .AND. x<=lo) THEN
              WRITE (*,90100) x, lo
              CYCLE TRY_LOOP
            END IF
          END IF LO_CHECK


!        Check upper bound if required.

90200     FORMAT ('The value entered > the upper bound allowed.'/ &
            'The value entered was: ',G20.7/'The upper bound is: ',G20.7)
90300     FORMAT ('The value entered >= the upper bound allowed.'/ &
            'The value entered was: ',G20.7/'The upper bound is: ',G20.7)

HI_CHECK: IF (hi_present) THEN
            IF (use_hi_eq_ok .AND. x>hi) THEN
              WRITE (*,90200) x, hi
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_hi_eq_ok .AND. x>=hi) THEN
              WRITE (*,90300) x, hi
              CYCLE TRY_LOOP
            END IF
          END IF HI_CHECK

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN


        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) 'Too many failed attempts at entry in &
            &get_double_scalar. ABORT'
          STOP 'Too many failed attempts at entry in get_double_scalar. &
            &ABORT'
        END IF

      END SUBROUTINE get_double_scalar

!*********************************************************************

      SUBROUTINE get_integer_array(format,lo,lo_eq_ok,hi,hi_eq_ok,x, &
          number_wanted,failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: number_wanted
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Array Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: hi(:), lo(:)
        INTEGER, INTENT (OUT) :: x(:)
! ..
! .. Local Scalars ..
        INTEGER :: check_hi_status, check_lo_status, i, ios, i_try, &
          use_n_wanted
        LOGICAL :: use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT, SIZE
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

!!!   CHECK_[LO|HI]_STATUS
!!!   -1 No check
!!!    1 Check against first and only element in bounds array
!!!   >1 Check against corresponding element in bounds array

        check_lo_status = -1
        IF (PRESENT(lo)) THEN
          check_lo_status = 1
          IF (SIZE(lo)/=1) check_lo_status = 2
        END IF

        check_hi_status = -1
        IF (PRESENT(hi)) THEN
          check_hi_status = 1
          IF (SIZE(hi)/=1) check_hi_status = 2
        END IF

        IF (PRESENT(number_wanted)) THEN
          use_n_wanted = number_wanted
        ELSE
          use_n_wanted = SIZE(x)
        END IF

TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again from the beginning.'

          IF (PRESENT(format)) WRITE (*,format)

          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) (x(i),i=1,use_n_wanted)
          IF (ios/=0) THEN
            WRITE (*,*) &
              'Some entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

CHECK_VALUES: DO i = 1, use_n_wanted

!        Check lower bound if required.

90000       FORMAT ('The ',I3, &
              '''th value entered < the lower bound allowed.'/ &
              'The value entered was: ',I20/'The lower bound is: ',I20)
90100       FORMAT ('The ',I3, &
              '''th value entered <= the lower bound allowed.'/ &
              'The value entered was: ',I20/'The lower bound is: ',I20)

LO_CHECK:   IF (check_lo_status>0) THEN
              IF (check_lo_status==1) THEN
                IF (use_lo_eq_ok .AND. x(i)<lo(1)) THEN
                  WRITE (*,90000) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(1)) THEN
                  WRITE (*,90100) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_lo_eq_ok .AND. x(i)<lo(i)) THEN
                  WRITE (*,90000) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(i)) THEN
                  WRITE (*,90100) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF LO_CHECK


!        Check upper bound if required.

90200       FORMAT ('The ',I3, &
              '''th value entered > the upper bound allowed.'/ &
              'The value entered was: ',I20/'The upper bound is: ',I20)
90300       FORMAT ('The ',I3, &
              '''th value entered >= the upper bound allowed.'/ &
              'The value entered was: ',I20/'The upper bound is: ',I20)

HI_CHECK:   IF (check_hi_status>0) THEN
              IF (check_hi_status==1) THEN
                IF (use_hi_eq_ok .AND. x(i)>hi(1)) THEN
                  WRITE (*,90200) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(1)) THEN
                  WRITE (*,90300) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_hi_eq_ok .AND. x(i)>hi(i)) THEN
                  WRITE (*,90200) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(i)) THEN
                  WRITE (*,90300) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF HI_CHECK

          END DO CHECK_VALUES

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN


        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) 'Too many failed attempts at entry in &
            &get_integer_array. ABORT'
          STOP 'Too many failed attempts at entry in get_integer_array. &
            &ABORT'
        END IF

      END SUBROUTINE get_integer_array

!*********************************************************************

      SUBROUTINE get_integer_scalar(format,lo,lo_eq_ok,hi,hi_eq_ok,x, &
          failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: hi, lo
        INTEGER, INTENT (OUT) :: x
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Local Scalars ..
        INTEGER :: ios, i_try
        LOGICAL :: hi_present, lo_present, use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

        lo_present = PRESENT(lo)

        hi_present = PRESENT(hi)


TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again.'

          IF (PRESENT(format)) WRITE (*,format)

          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) x
          IF (ios/=0) THEN
            WRITE (*,*) 'Entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

!        Check lower bound if required.

90000     FORMAT ('The value entered < the lower bound allowed.'/ &
            'The value entered was: ',I20/'The lower bound is: ',I20)
90100     FORMAT ('The value entered <= the lower bound allowed.'/ &
            'The value entered was: ',I20/'The lower bound is: ',I20)

LO_CHECK: IF (lo_present) THEN
            IF (use_lo_eq_ok .AND. x<lo) THEN
              WRITE (*,90000) x, lo
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_lo_eq_ok .AND. x<=lo) THEN
              WRITE (*,90100) x, lo
              CYCLE TRY_LOOP
            END IF
          END IF LO_CHECK


!        Check upper bound if required.

90200     FORMAT ('The value entered > the upper bound allowed.'/ &
            'The value entered was: ',I20/'The upper bound is: ',I20)
90300     FORMAT ('The value entered >= the upper bound allowed.'/ &
            'The value entered was: ',I20/'The upper bound is: ',I20)

HI_CHECK: IF (hi_present) THEN
            IF (use_hi_eq_ok .AND. x>hi) THEN
              WRITE (*,90200) x, hi
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_hi_eq_ok .AND. x>=hi) THEN
              WRITE (*,90300) x, hi
              CYCLE TRY_LOOP
            END IF
          END IF HI_CHECK

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN

        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) 'Too many failed attempts at entry in &
            &get_integer_scalar. ABORT'
          STOP 'Too many failed attempts at entry in get_integer_scalar. &
            &ABORT'
        END IF

      END SUBROUTINE get_integer_scalar

!*********************************************************************

      SUBROUTINE get_real_array(format,lo,lo_eq_ok,hi,hi_eq_ok,x, &
          number_wanted,failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
        INTEGER, PARAMETER :: rlkind = KIND(0.0)
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: number_wanted
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Array Arguments ..
        REAL (rlkind), OPTIONAL, INTENT (IN) :: hi(:), lo(:)
        REAL (rlkind), INTENT (OUT) :: x(:)
! ..
! .. Local Scalars ..
        INTEGER :: check_hi_status, check_lo_status, i, ios, i_try, &
          use_n_wanted
        LOGICAL :: use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC KIND, PRESENT, SIZE
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

!!!   CHECK_[LO|HI]_STATUS
!!!   -1 No check
!!!    1 Check against first and only element in bounds array
!!!   >1 Check against corresponding element in bounds array

        check_lo_status = -1
        IF (PRESENT(lo)) THEN
          check_lo_status = 1
          IF (SIZE(lo)/=1) check_lo_status = 2
        END IF

        check_hi_status = -1
        IF (PRESENT(hi)) THEN
          check_hi_status = 1
          IF (SIZE(hi)/=1) check_hi_status = 2
        END IF

        IF (PRESENT(number_wanted)) THEN
          use_n_wanted = number_wanted
        ELSE
          use_n_wanted = SIZE(x)
        END IF

TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again from the beginning.'

          IF (PRESENT(format)) WRITE (*,format)

          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) (x(i),i=1,use_n_wanted)
          IF (ios/=0) THEN
            WRITE (*,*) &
              'Some entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

CHECK_VALUES: DO i = 1, use_n_wanted

!        Check lower bound if required.

90000       FORMAT ('The ',I3, &
              '''th value entered < the lower bound allowed.'/ &
              'The value entered was: ',G20.7/'The lower bound is: ', &
              G20.7)
90100       FORMAT ('The ',I3, &
              '''th value entered <= the lower bound allowed.'/ &
              'The value entered was: ',G20.7/'The lower bound is: ', &
              G20.7)

LO_CHECK:   IF (check_lo_status>0) THEN
              IF (check_lo_status==1) THEN
                IF (use_lo_eq_ok .AND. x(i)<lo(1)) THEN
                  WRITE (*,90000) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(1)) THEN
                  WRITE (*,90100) i, x(i), lo(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_lo_eq_ok .AND. x(i)<lo(i)) THEN
                  WRITE (*,90000) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_lo_eq_ok .AND. x(i)<=lo(i)) THEN
                  WRITE (*,90100) i, x(i), lo(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF LO_CHECK


!        Check upper bound if required.

90200       FORMAT ('The ',I3, &
              '''th value entered > the upper bound allowed.'/ &
              'The value entered was: ',G20.7/'The upper bound is: ', &
              G20.7)
90300       FORMAT ('The ',I3, &
              '''th value entered >= the upper bound allowed.'/ &
              'The value entered was: ',G20.7/'The upper bound is: ', &
              G20.7)

HI_CHECK:   IF (check_hi_status>0) THEN
              IF (check_hi_status==1) THEN
                IF (use_hi_eq_ok .AND. x(i)>hi(1)) THEN
                  WRITE (*,90200) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(1)) THEN
                  WRITE (*,90300) i, x(i), hi(1)
                  CYCLE TRY_LOOP
                END IF
              ELSE
                IF (use_hi_eq_ok .AND. x(i)>hi(i)) THEN
                  WRITE (*,90200) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                ELSE IF ( .NOT. use_hi_eq_ok .AND. x(i)>=hi(i)) THEN
                  WRITE (*,90300) i, x(i), hi(i)
                  CYCLE TRY_LOOP
                END IF
              END IF
            END IF HI_CHECK

          END DO CHECK_VALUES

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN


        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) &
            'Too many failed attempts at entry in get_real_array. ABORT'
          STOP &
            'Too many failed attempts at entry in get_real_array. ABORT'
        END IF

      END SUBROUTINE get_real_array

!*********************************************************************

      SUBROUTINE get_real_scalar(format,lo,lo_eq_ok,hi,hi_eq_ok,x,failed)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: max_try = 3
        INTEGER, PARAMETER :: spkind = KIND(0.0)
! ..
! .. Scalar Arguments ..
        REAL (spkind), OPTIONAL, INTENT (IN) :: hi, lo
        REAL (spkind), INTENT (OUT) :: x
        LOGICAL, OPTIONAL, INTENT (OUT) :: failed
        LOGICAL, OPTIONAL, INTENT (IN) :: hi_eq_ok, lo_eq_ok
        CHARACTER (*), OPTIONAL, INTENT (IN) :: format
! ..
! .. Local Scalars ..
        INTEGER :: ios, i_try
        LOGICAL :: hi_present, lo_present, use_hi_eq_ok, use_lo_eq_ok
! ..
! .. Intrinsic Functions ..
        INTRINSIC KIND, PRESENT
! ..
! .. Executable Statements ..

!     Query user.
        use_lo_eq_ok = .TRUE.
        IF (PRESENT(lo_eq_ok)) use_lo_eq_ok = lo_eq_ok

        use_hi_eq_ok = .TRUE.
        IF (PRESENT(hi_eq_ok)) use_hi_eq_ok = hi_eq_ok

        lo_present = PRESENT(lo)

        hi_present = PRESENT(hi)


TRY_LOOP: DO i_try = 1, max_try

          IF (i_try>1) WRITE (*,*) 'Please try again.'

          IF (PRESENT(format)) WRITE (*,format)

          CALL prompt

!     Read in value.

          READ (*,*,iostat=ios) x
          IF (ios/=0) THEN
            WRITE (*,*) 'Entered value not accepted as a legal number.'
            CYCLE TRY_LOOP
          END IF

!     Check bound on each entry in X.

!        Check lower bound if required.

90000     FORMAT ('The value entered < the lower bound allowed.'/ &
            'The value entered was: ',G20.7/'The lower bound is: ',G20.7)
90100     FORMAT ('The value entered <= the lower bound allowed.'/ &
            'The value entered was: ',G20.7/'The lower bound is: ',G20.7)

LO_CHECK: IF (lo_present) THEN
            IF (use_lo_eq_ok .AND. x<lo) THEN
              WRITE (*,90000) x, lo
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_lo_eq_ok .AND. x<=lo) THEN
              WRITE (*,90100) x, lo
              CYCLE TRY_LOOP
            END IF
          END IF LO_CHECK


!        Check upper bound if required.

90200     FORMAT ('The value entered > the upper bound allowed.'/ &
            'The value entered was: ',G20.7/'The upper bound is: ',G20.7)
90300     FORMAT ('The value entered >= the upper bound allowed.'/ &
            'The value entered was: ',G20.7/'The upper bound is: ',G20.7)

HI_CHECK: IF (hi_present) THEN
            IF (use_hi_eq_ok .AND. x>hi) THEN
              WRITE (*,90200) x, hi
              CYCLE TRY_LOOP
            ELSE IF ( .NOT. use_hi_eq_ok .AND. x>=hi) THEN
              WRITE (*,90300) x, hi
              CYCLE TRY_LOOP
            END IF
          END IF HI_CHECK

          IF (PRESENT(failed)) failed = .FALSE.
          RETURN


        END DO TRY_LOOP

!!! Failure if we reach here

        IF (PRESENT(failed)) THEN
          failed = .TRUE.
          RETURN
        ELSE
          WRITE (*,*) 'Too many failed attempts at entry in &
            &get_real_scalar. ABORT'
          STOP 'Too many failed attempts at entry in get_real_scalar. &
            &ABORT'
        END IF

      END SUBROUTINE get_real_scalar

!*********************************************************************

      SUBROUTINE prompt
! Writes a prompt to module unit STDOUT, without advancing the line.
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Executable Statements ..

        WRITE (stdout,'(1X,''> '')',advance='NO')

        RETURN

      END SUBROUTINE prompt

!*********************************************************************

    END MODULE get_numbers_mod
