    MODULE open_file_mod
! .. Implicit None Statement ..
      IMPLICIT NONE
! ..
! .. Default Accessibility ..
      PRIVATE
! ..
! .. Parameters ..
      INTEGER, PARAMETER :: stdin = 5, stdout = 6
! ..
! .. Public Statements ..
      PUBLIC :: open_file, open_named_file
! ..
    CONTAINS

!*********************************************************************

      FUNCTION get_free_unit_number()
!!!-------------------------------------------------------------------
!!!                        Get Free Unit Number
!!!
!!! DESCRIPTION:
!!!
!!! Returns an existing FORTRAN unit number that is not in current
!!! use, has a value between unit_min and unit_max inclusive, and is
!!! not equal to stdin or stdout.  Returns -1 if unable to find such a
!!! unit number.
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Function Return Value ..
        INTEGER :: get_free_unit_number
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: unit_max = 99, unit_min = 10
! ..
! .. Local Scalars ..
        INTEGER, SAVE :: last_unit = unit_min
        LOGICAL :: exists, opened
! ..
! .. Executable Statements ..

        get_free_unit_number = last_unit

GET_UNIT: DO

          get_free_unit_number = get_free_unit_number + 1

          IF (get_free_unit_number==stdin .OR. get_free_unit_number== &
            stdout) CYCLE GET_UNIT

          IF (get_free_unit_number>unit_max) get_free_unit_number = &
            unit_min

          INQUIRE (unit=get_free_unit_number,exist=exists,opened=opened)

          IF (exists .AND. .NOT. opened) EXIT GET_UNIT

          IF (get_free_unit_number==last_unit) THEN

            get_free_unit_number = -1

            RETURN

          END IF

        END DO GET_UNIT

        last_unit = get_free_unit_number

        RETURN

      END FUNCTION get_free_unit_number

!*********************************************************************

      SUBROUTINE open_file(mssg,unit,read,name,err,confirm,append_ok, &
          delim)
! ----------------------------------------------------------------------
! Prompts use for file name using mssg and opens file.  Returns
! unit and (optionally) name.
!                             ARGUMENTS
! mssg: A  Fortran format that determines the  prompting message to
! be displayed to the user.
! CHARACTER(LEN=*), INTENT(IN) :: mssg
! unit:  The unit  number  of the  opened  file if  the routine  is
! successful.  If the routine is unsuccessful returns -1.
! INTEGER, INTENT(OUT) :: unit
! read: Logical.  If .TRUE., user  intends  to read file; if  .FALSE.,
! user  intends    to  write   file.    All   files are   opened  with
! action='readwrite'.   This item is   only used to select the  user's
! choices when the  file exists.  For example, a  file to be read must
! exist.  If missing, defaults to .TRUE.
! LOGICAL, INTENT(IN), OPTIONAL :: read
! name: System file name of successfully opened file (on most systems,
! this   includes the full  path).  If  not present,  this name is not
! returned.
! CHARACTER(LEN=*), INTENT(OUT), OPTIONAL :: name
! err: Completion  status.   0  ok  others  signal error  as  per  the
! following table.   If not present,  routine aborts unless the status
! is 0.
!  -1 - no units available ( or left )
!   0 - no error, file is open with name name and unit unit
!       if read is .TRUE., file exists and is positioned at begin
!       if read is .FALSE., file exists and is ready to write
!       at the position specified by user if applicable
!   1 - too many failures at file specification
!   2 - user chose to abort by choosing (q) or entering 'quit' instead
!       of file name
!   3 - user chose to go back by entering 'back' instead of file name
! INTEGER, INTENT(OUT), OPTIONAL :: err
! confirm:   Logical.    IF  .TRUE.,  asks      user to confirm   file
! specification after OPEN succeeds.  If .FALSE., does not ask user to
! confirm.  If missing defaults to .FALSE.
! LOGICAL, INTENT(IN), OPTIONAL :: confirm
! append_ok: Logical.  IF .TRUE.   and read is .FALSE.,  appending to
! the file  is allowed if  the file exists.   If  .FALSE., it  is not.
! Rarely used.  If missing defaults to .TRUE.
! LOGICAL, INTENT(IN), OPTIONAL :: append_ok
! delim: Used  for the delim= option for  open.  This is the character
! that delimits strings and causes those  characters within the string
! to be doubled  in  list-directed (* format)  and  namelist-formatted
! output (write).  It  is ignored for formatted  input (read).  Rarely
! used.  If not present, the value 'none'  is used.  Legal values are:
! 'none', 'apostrophe',  or 'quote', but its  value  is not checked by
! this routine.
! CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: delim
! ----------------------------------------------------------------------
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (OUT) :: err
        INTEGER, INTENT (OUT) :: unit
        LOGICAL, OPTIONAL, INTENT (IN) :: append_ok, confirm, read
        CHARACTER (*), OPTIONAL, INTENT (IN) :: delim
        CHARACTER (*), INTENT (IN) :: mssg
        CHARACTER (*), OPTIONAL, INTENT (OUT) :: name
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: local_name_len = 1000, max_bad = 3
! ..
! .. Local Scalars ..
        INTEGER :: action_code, i, ios, j, m, num_action_chars
        LOGICAL :: delete_ok, local_append_ok, local_confirm, local_read, &
          qerr, qexists, qopened, qyes
        CHARACTER (4) :: action_chars, command
        CHARACTER (1000) :: action_mssg
        CHARACTER (10) :: local_delim
        CHARACTER (local_name_len) :: local_name
        CHARACTER (7) :: rw_status
! ..
! .. Intrinsic Functions ..
        INTRINSIC ADJUSTL, IACHAR, LEN_TRIM, PRESENT, TRIM
! ..
! .. Executable Statements ..

!!! Set defaults for OPTIONAL INTENT(IN) args
        local_read = .TRUE.
        IF (PRESENT(read)) local_read = read

        local_confirm = .FALSE.
        IF (PRESENT(confirm)) local_confirm = confirm

        local_append_ok = .TRUE.
        IF (PRESENT(append_ok)) local_append_ok = append_ok

        local_delim = 'none'
        IF (PRESENT(delim)) local_delim = delim

!!! Get free unit number or bomb out

        unit = get_free_unit_number()

        IF (unit<0) THEN

          CALL handle_error(-1) ! unable to get free unit number

          RETURN

        END IF

!!! Get a file name from the user

GET_NAME: DO i = 1, max_bad

          IF (i>1) WRITE (stdout,'(/5X,''Please try again ...'')')

!!! Tell the user what the name is for

          WRITE (stdout,mssg)

          CALL prompt

          READ (stdin,'(A)',iostat=ios) local_name

          IF (ios/=0) THEN

            WRITE (stdout,'(/5X,''Could not read the file name.'')')

            CYCLE GET_NAME

          END IF

          IF (local_name==' ') THEN

            WRITE (stdout,'(/5X,''A blank name is not allowed.'')')

            CYCLE GET_NAME

          END IF

          local_name = ADJUSTL(local_name)

!!! Check for unusual characters

          DO m = 1, LEN_TRIM(local_name)

            SELECT CASE (IACHAR(local_name(m:m)))

            CASE (:31,127:)

              WRITE (stdout,'(/5X,''Unusual character(s) were &
                &detected in the file name:''/5X,A/5X,''This &
                &is probably a typographical error.'')') TRIM(local_name)

              CALL get_yn('(/5X,''Are you sure you want to &
                &use this file name? [yes or no]'')',qyes,qerr)

              IF (qerr) CYCLE GET_NAME

              IF (qyes) EXIT

              CYCLE GET_NAME

            END SELECT

          END DO

!!! Check for "back" or "quit"

          IF (LEN_TRIM(local_name)==4) THEN

            command = local_name(1:4)

            CALL low_case_string(command)

BACK_OR_QUIT: SELECT CASE (command)

            CASE ('back') BACK_OR_QUIT

              CALL handle_error(3) ! user entered 'back'

              RETURN

            CASE ('quit') BACK_OR_QUIT

              CALL handle_error(2) ! user quit

              RETURN

            END SELECT BACK_OR_QUIT

          END IF

!!! Get system version of the file name or return error

          INQUIRE (file=local_name,name=local_name,iostat=ios)

          IF (ios/=0) THEN

            WRITE (stdout,'(/5X,''Could not get the system &
              &version of the file name.''/5X,''It is probably &
              &a bad file name.'')')

            CYCLE GET_NAME

          END IF

!!! Find out if it exists

          INQUIRE (file=local_name,exist=qexists,iostat=ios)

          IF (ios/=0) THEN

            WRITE (stdout,'(/5X,''Could not determine if the &
              &file is available.''/5X,''It is probably a bad &
              &file name.'')')

            CYCLE GET_NAME

          END IF

!!! Find out if file is already in use.

          INQUIRE (file=local_name,opened=qopened,iostat=ios)

          IF (ios/=0) THEN

            WRITE (stdout,'(/5X,''Could not determine if the &
              &file is already in use.''/5X,''It is probably &
              &a bad file name.'')')

            CYCLE GET_NAME

          END IF

          IF (qopened) THEN

            WRITE (stdout, &
              '(/5X,''The file: ''/5X,A/5X,''is already in use.'')') &
              TRIM(local_name)

            CYCLE GET_NAME

          END IF

!!! Try to open the file in the desired mode.

          IF (local_read) THEN

!!! The file is to be opened for reading and must exist

            IF (qexists) THEN

!!! Open file for reading (and associate with unit)

              OPEN (unit=unit,file=local_name,status='OLD', &
                position='REWIND',action='READWRITE',iostat=ios, &
                delim=local_delim)

              IF (ios/=0) THEN !!! error

                WRITE (stdout,'(/5X,''Could not open the file:''/5X,A/5X,'&
                  &'for reading.'')') TRIM(local_name)

                CYCLE GET_NAME

              END IF

!!! Set up confirmation information.

              action_mssg = '(/5X,''Opened file:''/5X,''' // &
                TRIM(local_name) // '''/5X,''for reading.'''

              delete_ok = .FALSE.

!!! Fall through to confirmation code at end of GET_NAME

            ELSE

!!! File does not exist, tell user, try again.

              WRITE (stdout, &
                '(/5X,''The file:''/5X,A/5X,''is not available.'')') &
                TRIM(local_name)

              CYCLE GET_NAME

            END IF

          ELSE

!!! File is to be opened for writing.

            IF (qexists) THEN

!!! File exists; set up choices allowed for the user.

              action_mssg = '(/5X,''The file:''/5X,''' // &
                TRIM(local_name) // '''/5X,''exists.''//5X,''Choose &
                &an action by entering the corresponding letter:''/ &
                & 7X,''(q)uit file specification'

              IF ( .NOT. PRESENT(err)) action_mssg((LEN_TRIM( &
                action_mssg)+1):) = ' (this will abort the program)'

              action_mssg((LEN_TRIM(action_mssg)+1):) = '''/ &
                & 7X,''(r)etry file specification (if you made &
                &a mistake)''/  7X,''(o)verwrite file (this &
                &will delete the existing file)'')'

              IF (local_append_ok) THEN

                action_mssg(LEN_TRIM(action_mssg):) &
                  = '/  7X,''(a)ppend to file'')'

                action_chars = 'qroa'

                num_action_chars = 4

              ELSE

                action_chars = 'qro'

                num_action_chars = 3

              END IF

!!! Ask user for appropriate action.

GET_ACTION:   DO j = 1, max_bad

                IF (j>1) WRITE (stdout, &
                  '(/5X,''Please choose a different action ...'')')

                CALL get_character(action_mssg,action_chars, &
                  num_action_chars,action_code,qerr)

                IF (qerr) THEN !!! No valid response; try specification again

                  WRITE (stdout, &
                    '(/5X,''Unable to get a valid choice of action.'')')

                  CYCLE GET_NAME

                END IF

ACTION_CHOICE:  SELECT CASE (action_code)

                CASE (1) ACTION_CHOICE !!! Quit file specification

                  CALL handle_error(2) ! user quit

                  RETURN

                CASE (2) ACTION_CHOICE !!! Retry file specification

                  CYCLE GET_NAME

                CASE (3) ACTION_CHOICE !!! Overwrite file ...

                  OPEN (unit=unit,file=local_name,status='REPLACE', &
                    position='REWIND',action='READWRITE',iostat=ios, &
                    delim=local_delim)

                  IF (ios==0) THEN !!! opened without error

!!! Set up information to confirm overwrite

                    action_mssg = '(/5X,''Opened file:''/5X,''' // &
                      TRIM(local_name) // '''/5X,''for overwriting &
                      &(the existing file has been deleted).'''

                    delete_ok = .TRUE.

!!! Fall through to write check code

                    EXIT GET_ACTION

                  END IF

!!! Error opening, tell user, try another action

                  WRITE (stdout, &
                    '(/5X,''Could not overwrite file:''/5X,A)') &
                    TRIM(local_name)

                CASE (4) ACTION_CHOICE !!! Append to file ...

                  OPEN (unit=unit,file=local_name,status='OLD', &
                    position='APPEND',action='READWRITE',iostat=ios, &
                    delim=local_delim)

                  IF (ios==0) THEN !!! opened without error

!!! Set up information to confirm append

                    action_mssg = '(/5X,''Opened file:''/5X,''' // &
                      TRIM(local_name) // '''/5X,''for appending.'''

                    delete_ok = .FALSE.

!!! Fall through to write check code

                    EXIT GET_ACTION

                  END IF

!!! Error opening, tell user, try another action

                  WRITE (stdout, &
                    '(/5X,''Could not append to file:''/5X,A)') &
                    TRIM(local_name)

                END SELECT ACTION_CHOICE

              END DO GET_ACTION

            ELSE

!!! File for writing does not exist.  Create it.

              OPEN (unit=unit,file=local_name,status='NEW', &
                action='READWRITE',iostat=ios,delim=local_delim, &
                position='REWIND')

              IF (ios/=0) THEN !!! error

                WRITE (stdout,'(/5X,''Could not create file:''/5X,A/5X,''f&
                  &or writing.'')') TRIM(local_name)

                CYCLE GET_NAME

              END IF

!!! Set up confirmation information

              action_mssg = '(/5X,''Created file:''/5X,''' // &
                TRIM(local_name) // '''/5X,''for writing.''/'

              delete_ok = .TRUE.

            END IF

!!! Check writability of files opened for writing

            INQUIRE (unit=unit,write=rw_status)

            IF (rw_status/='YES') THEN

              action_mssg((LEN_TRIM(action_mssg)+1):) &
                = '/5X,''However, it cannot be written to.'')'

              WRITE (stdout,action_mssg)

!!! close or delete as appropriate

              CALL close_or_new_unit

              IF (unit>=0) CYCLE GET_NAME

              CALL handle_error(-1) ! unable to get free unit number

              RETURN

            END IF

          END IF

!!! Check readability of all files

          INQUIRE (unit=unit,read=rw_status)

          IF (rw_status/='YES') THEN

            action_mssg((LEN_TRIM(action_mssg)+1):) &
              = '/5X,''However, it cannot be read.'')'

            WRITE (stdout,action_mssg)

!!! close or delete as appropriate

            CALL close_or_new_unit

            IF (unit>=0) CYCLE GET_NAME

            CALL handle_error(-1) ! unable to get free unit number

            RETURN

          END IF

CONFIRM_CODE: IF (local_confirm) THEN

!!! Confirmation code begins.  Should only get here after a successful
!!! open.

!!! add confirmation choices to message format; get user's choice.

            action_mssg((LEN_TRIM(action_mssg)+1):) = '//5X,''Choose &
              &an action by entering the corresponding letter:''/ &
              & 7X,''(q)uit file specification'

            IF ( .NOT. PRESENT(err)) action_mssg((LEN_TRIM( &
              action_mssg)+1):) = ' (this will abort the program)'

            action_mssg((LEN_TRIM(action_mssg)+1):) = '''/ &
              & 7X,''(r)etry file specification (if you made &
              &a mistake)''/  7X,''(p)roceed'')'

            CALL get_character(action_mssg,'qrp',3,action_code,qerr)

            IF (qerr) THEN !!! No valid response; try specification again

              WRITE (stdout, &
                '(/5X,''Unable to get a valid choice of action.'')')

              action_code = 2

            END IF

CONFIRM_CHOICE: SELECT CASE (action_code)

            CASE (1) CONFIRM_CHOICE !!! Quit file specification

              CALL close_or_new_unit

              CALL handle_error(2) ! user quit

              RETURN

            CASE (2) CONFIRM_CHOICE !!! Retry file specification

!!! close or delete as appropriate

              CALL close_or_new_unit

              IF (unit>=0) CYCLE GET_NAME

              CALL handle_error(-1) ! unable to get free unit number

              RETURN

            CASE (3) CONFIRM_CHOICE !!! Proceed; fall through

            END SELECT CONFIRM_CHOICE

          END IF CONFIRM_CODE

!!! SUCCESSFUL RETURN -- should only get here if successful

!!! set OPTIONAL INTENT(OUT) args

          IF (PRESENT(err)) err = 0

          IF (PRESENT(name)) name = local_name

          RETURN

        END DO GET_NAME

        CALL handle_error(1) ! too many bad file specifications

        RETURN

      CONTAINS

!.....................................................................

        SUBROUTINE close_or_new_unit
!!!-------------------------------------------------------------------
!!!
!!! If ( DELETE_OK   ), tries   to close  the  unit   and delete  the
!!! associated  file.  Returns if  this  succeeds; if it fails,  tells
!!! user and ( or  if ( .NOT. DELETE_OK  ) ) tries an ordinary  close.
!!! Returns if this succeeds; if it fails, tells user and assigns UNIT
!!! = GET_FREE_UNIT_NUMBER( ).   OPEN_FILE  should check the value  of
!!! UNIT on return.
!!!
!!! Uses host-associated symbols:
!!!   values used:    DELETE_OK,   LOCAL_NAME,  UNIT,  STDOUT
!!!   values set :    UNIT
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Local Scalars ..
          INTEGER :: sub_ios
! ..
! .. Executable Statements ..

!!! Try to delete file if indicated
          IF (delete_ok) THEN

            CLOSE (unit,status='DELETE',iostat=sub_ios)

            IF (sub_ios==0) RETURN !!! Closed without error

            WRITE (stdout,'(/5X,''Could not delete file:''/5X,A)') &
              TRIM(local_name)

          END IF

          CLOSE (unit,iostat=sub_ios)

          IF (sub_ios==0) RETURN !!! Closed without error

          WRITE (stdout,'(/5X,''Could not close file:''/5X,A/5X,''Warning:&
            & Possible system error'')') TRIM(local_name)

          unit = get_free_unit_number()

        END SUBROUTINE close_or_new_unit

!.....................................................................

        SUBROUTINE get_character(mssg,chars,num_chars,which,qerr)
!!!-------------------------------------------------------------------
!!!
!!! Displays  MSSG   to  user  (explaining  character   choices), gets
!!! response from user,  evaluates first non-blank character,  returns
!!! index of response character in CHARS(1:NUM_CHARS) as WHICH.
!!!
!!! CHARS should all be lower case.  Response  letter is translated to
!!! lower case before evaluation.
!!!
!!! If  no good answer in MAX_BAD  tries, QERR == TRUE, otherwise QERR
!!! == FALSE.
!!!
!!! NOTE: if NUM_CHARS  is  not positive  and  at most  the  length of
!!! CHARS, undefined behavior may result.
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Parameters ..
          INTEGER, PARAMETER :: line_len = 80, max_bad = 3
! ..
! .. Scalar Arguments ..
          INTEGER, INTENT (IN) :: num_chars
          INTEGER, INTENT (OUT) :: which
          LOGICAL, INTENT (OUT) :: qerr
          CHARACTER (*), INTENT (IN) :: chars, mssg
! ..
! .. Local Scalars ..
          INTEGER :: i, ios, j
          CHARACTER (1) :: answer
          CHARACTER (line_len) :: line
! ..
! .. Intrinsic Functions ..
          INTRINSIC INDEX
! ..
! .. Executable Statements ..

          qerr = .FALSE.

GET_ANSWER: DO i = 1, max_bad

            IF (i>1) WRITE (stdout,'(/5X,''Please try again ...'')')

            WRITE (stdout,mssg)

            WRITE (stdout,'(1X,''Please enter one of ['',A,'']:'')', &
              advance='NO') chars(1:num_chars)

            CALL prompt

            READ (stdin,'(A)',iostat=ios) line

            IF (ios/=0) THEN

              WRITE (stdout,'(/5X,''Could not read the response.'')')

              CYCLE GET_ANSWER

            END IF

            DO j = 1, line_len

              IF (line(j:j)/=' ') EXIT

            END DO

            IF (j>line_len) THEN

              WRITE (stdout,'(/5X,''A blank line is not allowed.'')')

              CYCLE GET_ANSWER

            END IF

            answer = low_case(line(j:j))

            which = INDEX(chars(1:num_chars),answer)

            IF (which>0) RETURN

            WRITE (stdout,'(/5X,A,'' is not allowed.'')') answer

          END DO GET_ANSWER

          qerr = .TRUE.

          RETURN

        END SUBROUTINE get_character

!.....................................................................

        SUBROUTINE get_yn(mssg,qyes,qerr)
!!!-------------------------------------------------------------------
!!!
!!! Uses MSSG  to elicit a   Y/N response  from   the user by  calling
!!! GET_CHARACTER.  If  GET_CHARACTER  fails, QERR == TRUE;  otherwise
!!! QERR == FALSE.
!!!
!!! The value   for QGETYN is  undefined  if QERR ==   TRUE; otherwise
!!! QGETYN is TRUE if  the  user enters a  'y', or  FALSE if the  user
!!! enters an 'n'.
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Scalar Arguments ..
          LOGICAL, INTENT (OUT) :: qerr, qyes
          CHARACTER (*), INTENT (IN) :: mssg
! ..
! .. Local Scalars ..
          INTEGER :: which
! ..
! .. Executable Statements ..

          CALL get_character(mssg,'yn',2,which,qerr)

          IF ( .NOT. qerr) qyes = (which==1)

          RETURN

        END SUBROUTINE get_yn

!.....................................................................

        SUBROUTINE handle_error(code)
! --------------------------------------------------------------------
! Handles errors in open_file.
! If err is  present in call to host,  sets err =  code, unit = -1 and
! returns.
! If  err  is not present   in call to   host, writes error message to
! stdout and stops execution of the program.
!                                ARGUMENT
! code:  An integer error code to be handled.
! Legal values for code and their meanings:
!  -1 - no units available ( or left )
!   1 - too many failures at file specification
!   2 - user chose to abort by choosing (q) or entering 'quit' instead
!       of file name
!   3 - user chose to go back by entering 'back' instead of file name
! INTEGER, INTENT(IN) :: code
! Uses err, an optional argument  accessible  to this routine by  host
! association.
! --------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Scalar Arguments ..
          INTEGER, INTENT (IN) :: code
! ..
! .. Executable Statements ..

          IF (PRESENT(err)) THEN

            err = code

            unit = -1

            RETURN

          END IF

          SELECT CASE (code)

          CASE (-1) ! no unit

            WRITE (stdout,*) 'In subroutine open_file: Unable &
              &to get free unit number.  Abort!'

            STOP 'In subroutine open_file: Unable to get free &
              &unit number.  Abort!'

          CASE (1) ! too many failures

            WRITE (stdout,*) 'In subroutine open_file: Too &
              &many failed attempts at file specification. &
              & Abort!'

            STOP 'In subroutine open_file: Too many failed &
              &attempts at file specification.  Abort!'

          CASE (2) ! quit

            WRITE (stdout,*) 'In subroutine open_file: User &
              &aborted by entering ''quit'' instead of file &
              &name or choosing (q)uit file specification. &
              & Abort!'

            STOP 'In subroutine open_file: User aborted by &
              &entering ''quit'' instead of file name or choosing &
              &(q)uit file specification.  Abort!'

          CASE (3) ! back

            WRITE (stdout,*) 'In subroutine open_file: User &
              &aborted by entering ''back'' instead of file &
              &name.  Abort!'

            STOP 'In subroutine open_file: User aborted by &
              &entering ''back'' instead of file name.  Abort!'

          CASE DEFAULT ! bad code; catastrophic

            WRITE (stdout,*) 'In subroutine open_file: INTERNAL &
              &ERROR -- contained subroutine handle_error attempted &
              &to handle error code:', code, ' Abort!'

            STOP 'In subroutine open_file: INTERNAL ERROR -- &
              &contained subroutine handle_error attempted &
              &to handle bogus error code.  Abort!'

          END SELECT

        END SUBROUTINE handle_error

!.....................................................................

        FUNCTION low_case(chr)
!!!-------------------------------------------------------------------
!!!
!!! NAME: translate one character to LOWer CASE
!!!
!!!
!!! DESCRIPTION:
!!!
!!! If CHR is an upper case letter  'a..z' then returns the lower case
!!! letter corresponding.   If CHR is not an   upper case letter, then
!!! returns CHR.
!!!
!!!
!!! ARGUMENT:
!!!
!!! CHR --> Character to be translated to upper case.
!!!                              CHARACTER*1 CHR
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Function Return Value ..
          CHARACTER (1) :: low_case
! ..
! .. Parameters ..
          CHARACTER (26), PARAMETER :: locase = &
            'abcdefghijklmnopqrstuvwxyz'
          CHARACTER (26), PARAMETER :: upcase = &
            'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
! ..
! .. Scalar Arguments ..
          CHARACTER (1), INTENT (IN) :: chr
! ..
! .. Local Scalars ..
          INTEGER :: ix
          LOGICAL :: qnotuc
! ..
! .. Intrinsic Functions ..
          INTRINSIC INDEX, LGT, LLT
! ..
! .. Executable Statements ..

          qnotuc = (LLT(chr,'A') .OR. LGT(chr,'Z'))

          IF (qnotuc) THEN

            low_case = chr

            RETURN

          END IF

          ix = INDEX(upcase,chr)

          IF (ix<=0) THEN

            low_case = chr

          ELSE

            low_case = locase(ix:ix)

          END IF

          RETURN

        END FUNCTION low_case

!.....................................................................

        SUBROUTINE low_case_string(str)
!!!-------------------------------------------------------------------
!!!
!!! NAME: LOWer CASE a whole STRING
!!!
!!!
!!! DESCRIPTION:
!!!
!!! For each CHR in STR, If CHR is an upper case letter 'a..z' then it
!!! is replaced by the lower case letter corresponding.  If CHR is not
!!! an upper case letter, then nothing happens to CHR.
!!!
!!!
!!! ARGUMENT:
!!!
!!! STR <-> String to be translated to upper case.
!!!                             CHARACTER*(*) STR
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Parameters ..
          CHARACTER (26), PARAMETER :: locase = &
            'abcdefghijklmnopqrstuvwxyz'
          CHARACTER (26), PARAMETER :: upcase = &
            'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
! ..
! .. Scalar Arguments ..
          CHARACTER (*), INTENT (INOUT) :: str
! ..
! .. Local Scalars ..
          INTEGER :: ix, j
          LOGICAL :: qnotuc
          CHARACTER (1) :: chr
! ..
! .. Intrinsic Functions ..
          INTRINSIC INDEX, LEN_TRIM, LGT, LLT
! ..
! .. Executable Statements ..

          DO j = 1, LEN_TRIM(str)

            chr = str(j:j)

            qnotuc = (LLT(chr,'A') .OR. LGT(chr,'Z'))

            IF (qnotuc) CYCLE

            ix = INDEX(upcase,chr)

            IF (ix<=0) CYCLE

            str(j:j) = locase(ix:ix)

          END DO

          RETURN

        END SUBROUTINE low_case_string

!.....................................................................

        SUBROUTINE prompt
!!!-------------------------------------------------------------------
!!!
!!! Writes a prompt to STDOUT, without advancing the line.
!!!
!!!-------------------------------------------------------------------
! .. Implicit None Statement ..
          IMPLICIT NONE
! ..
! .. Executable Statements ..

          WRITE (stdout,'(1X,''> '')',advance='NO')

          RETURN

        END SUBROUTINE prompt

!.....................................................................

      END SUBROUTINE open_file

!*********************************************************************

      SUBROUTINE open_named_file(file_name,unit,read)
!!! Opens a file with name the contents of file_name
!!! This routine is for use with constructed file names
!!! Unit is the unit number of the file
!!! Aborts with an error message on any error
! .. Use Statements ..
        USE file_mod
! ..
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: stdout = 6
! ..
! .. Scalar Arguments ..
        INTEGER, INTENT (OUT) :: unit
        LOGICAL, INTENT (IN) :: read
        CHARACTER (*), INTENT (IN) :: file_name
! ..
! .. Local Scalars ..
        INTEGER :: ios
        LOGICAL :: exists, readable, writeable
! ..
! .. Intrinsic Functions ..
        INTRINSIC LEN_TRIM
! ..
! .. Executable Statements ..

!!! Default to bad unit
        unit = -1

        CALL file_exists(file_name,exists,readable,writeable)

        IF (read) THEN ! File to be opened for reading

          IF (exists .AND. readable) THEN

            unit = get_free_unit_number()

            OPEN (file=file_name,unit=unit,status='OLD', &
              position='REWIND',iostat=ios,action='READ')

            IF (ios==0) THEN

              RETURN

            ELSE

              IF ( .NOT. exists) WRITE (stdout,90000) &
                file_name(:LEN_TRIM(file_name))
              RETURN

90000         FORMAT (5X,'The following file does not exist, &
                &so could not be opened for reading:'/5X,A)


            END IF

            IF (exists .AND. .NOT. readable) WRITE (stdout,90100) &
              file_name(:LEN_TRIM(file_name))
            RETURN

90100       FORMAT (5X,'The following file exists but cannot &
              &be opened for reading:'/5X,A)


          END IF

        ELSE ! File to be written

          IF ( .NOT. exists .OR. writeable) THEN

            unit = get_free_unit_number()

            OPEN (file=file_name,unit=unit,position='REWIND',iostat=ios, &
              action='WRITE')

            IF (ios==0) THEN
              RETURN
            ELSE

              WRITE (stdout,90300) file_name(:LEN_TRIM(file_name))
              WRITE (stdout,90200) ios
90200         FORMAT (' Error code (iostat) : ',I10)
              RETURN

90300         FORMAT ( &
                ' The following file could not be opened for writing:'/ &
                5X,A)

            END IF

          ELSE ! File not writeable

            WRITE (stdout,90400) file_name(:LEN_TRIM(file_name))

90400       FORMAT (/5X, &
              'The following file exists and cannot be overwritten:'/5X, &
              A)
          END IF

        END IF

      END SUBROUTINE open_named_file

!*********************************************************************

    END MODULE open_file_mod
