    MODULE print_it
! Description:
! This module is intended to help automate the printing of user messages
! to the current output unit.  It is intended to be used with the output
! of format.specs,  a   Perl script.   It  should  be included with  USE
! statement(s) visible to all the code  which incorporates the output of
! format.specs.  Typically, this will   only require the  programmer  to
! assign values to SUB_STRINGS in order to produce attractive output.
! NOTE: The charcter strings in  SUB_STRINGS are processed to double any
! single  quotation marks  found,  and   the substitution  is   adjusted
! accordingly.  This should  make everything run  smoothly provided that
! the  string to be  substituted appears  as  it should appear on output
! (i.e.  the  quotation marks are  not pre-doubled), and that the format
! strings use  single   quotes  to  enclose character   constant  format
! elements.  If these conditions are not  met, the output may have twice
! the single  quotation marks as desired, but  the format  string should
! not cause an I/O error.
! Module Parameters:
! Note: If any of these three parameters  are changed, the corresponding
! variable in format.specs should probably be changed as well.
! mx_format_length -- The  maximum character length of a  format string.
!                     INTEGER, PARAMETER :: mx_format_length
! mx_subs          -- The maximum number of substitutions in each format
!                     string processed.
!                     INTEGER, PARAMETER :: mx_subs
! mx_sub_length    -- The maximum character length of  each substitution
!                     string.
!                     INTEGER, PARAMETER :: mx_sub_length
! Module Arrays:
! sub_pos          -- The indices of the  current format string at which
!                     the  substitutions  are  to  be  made.   The  I'th
!                     substitution  is made to the sub-string indexed by
!                     ( sub_pos(2*I-1) : sub_pos(2*I) ).
!                     INTEGER, DIMENSION(2*mx_subs) :: sub_pos
! sub_strings      -- The sub-strings  to be substituted into the format
!                     string. The I'th substitution gets sub_strings(I).
!      CHARACTER(LEN=mx_sub_length), DIMENSION( mx_subs ) :: sub_strings
! Module Scalars:
! always_print     -- Logical indicator of  under what circumstances the
!                     current format string should be printed.
!                        .TRUE.  -- Print it regarless of print_level.
!                        .FALSE. -- Print it according to print_level.
!                     Help messages generally set the value to .FALSE.
!                     LOGICAL :: always_print
! message_format   -- The 'default'  format string  to be  printed.  The
!                     Perl  script  format.specs  assigns  unnamed  text
!                     blocks to this character string,  so most messages
!                     will  probably  use  it.   It  doesn't need  to be
!                     declared in code using this module,  and it can be
!                     printed with a call to print_message_format().
!                     CHARACTER(LEN=mx_format_length) :: message_format
! num_subs         -- The number of substitutions to be preformed on the
!                     current format string.
!                     INTEGER :: num_subs
! print_level      -- The (user specified)  level at which  messages are
!                     to be printed.
!                     Allowable values of print_level:
!                        1 -- always print help
!                        2 -- ask whether help should be printed
!                        3 -- never print help
!                     INTEGER, SAVE :: print_level
! format_printed   -- Logical  value   indicating  whether  the  message
!                     format  was  printed  or  not.   This  is  set  by
!                     print_format( format_string ),  and is intended to
!                     help the calling routine  control the scrolling of
!                     messages,  particularly  in the  case of  optional
!                     messages   (where  always_print   ==   .FALSE.  ).
!                     LOGICAL, SAVE :: format_printed
! clear_screen_before_print -- logical, intended to be set by programmer.
! Module Subroutines:
! print_message_format()
!     calls print_format( message_format ).
! print_format( format_string )
!   CHARACTER(LEN=*), INTENT(INOUT) :: format_string
!     Prints  out format_string   using   the  above variables.    Calls
!     edit_format(  format_string  ),  and  sets  format_printed.   This
!     subroutine uses the complib routines GTCUIO and QYNGT.
! edit_message_format()
!     calls edit_format( message_format ).
! edit_format( format_string )
!   CHARACTER(LEN=*), INTENT(INOUT) :: format_string
!     Prepares format_string for output  by performing the substitutions
!     indicated.  It is called by print_format.
! for CLEAR_SCREEN( ), windowing routines, etc.
! .. Implicit None Statement ..
      IMPLICIT NONE
! ..
! .. Parameters ..
      INTEGER, PARAMETER :: mx_format_length = 5000, mx_subs = 100, &
        mx_sub_length = 100
! ..
! .. Local Scalars ..
      INTEGER :: num_subs
      INTEGER, SAVE :: print_level = 1, window_size = 24
      INTEGER, PRIVATE :: stdin = 5, stdout = 6
      LOGICAL :: always_print, print_off
      LOGICAL, SAVE :: clear_screen_before_print = .FALSE., &
        format_printed
      CHARACTER (mx_format_length) :: message_format
! ..
! .. Local Arrays ..
      INTEGER :: sub_pos(2*mx_subs)
      CHARACTER (mx_sub_length) :: sub_strings(mx_subs)
! ..
    CONTAINS

!*********************************************************************

      SUBROUTINE clear_screen
! writes out WINDOW_SIZE (in this module) blank lines
! to STDOUT
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Local Scalars ..
        INTEGER :: i
! ..
! .. Executable Statements ..

        DO i = 1, window_size

          WRITE (stdout,*)

        END DO

      END SUBROUTINE clear_screen

!*********************************************************************

      SUBROUTINE edit_format(format_string)
!                     EDIT_FORMAT( FORMAT_STRING )
! Substitutes the   first   NUM_SUBS elements of    SUB_STRINGS  for the
! sub-strings in FORMAT_STRING defined by (SUB_POS(2*I-1):SUB_POS(2*I)).
! Intended  to prepare FORMAT_STRING for output,  i.e.  after  a call to
! EDIT_FORMAT fills in the desired values,
!                    WRITE( stdout, FORMAT_STRING)
! should produce the desired message.
! NOTE: This code  assumes   that FORMAT_STRING uses single  quotes   to
! delimit the character strings, and therefore doubles any single quotes
! found in the SUB_STRINGS, adjusting the positions of the substitutions
! and copying FORMAT_STRING appropriately.
! Argument:
! FORMAT_STRING <-> A character string intended to be the template for a
!                   message.  On return the positions bounded by SUB_POS
!                   are altered.
!                     CHARACTER(LEN=*), INTENT(INOUT) :: FORMAT_STRING
! Module symbols used (see above):
! NUM_SUBS, SUB_POS, SUB_STRINGS
! .. Scalar argument ..
! .. Local scalar ..
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        CHARACTER (*), INTENT (INOUT) :: format_string
! ..
! .. Local Scalars ..
        INTEGER :: extra_fmt_chars, extra_sub_chars, i, jj
        CHARACTER (1) :: temp_char
        CHARACTER (mx_sub_length) :: temp_sub_string
! ..
! .. Executable Statements ..

        extra_fmt_chars = 0

SUBSTITUTE: DO i = 2, 2*num_subs, 2

! copy SUB_STRINGS( I/2 ) into TEMP_SUB_STRING,
! checking for single quotes along the way

          extra_sub_chars = 0

! loop over the significant characters in SUB_STRINGS( I/2 )

          DO jj = 1, sub_pos(i) - sub_pos(i-1) + 1

! copy this character

            temp_char = sub_strings(i/2) (jj:jj)

            temp_sub_string(jj+extra_sub_chars:jj+extra_sub_chars) &
              = temp_char

! if this character was a single quote, double it and keep track

            IF (temp_char=='''') THEN

! Advance TEMP_SUB_STRING index by one

              extra_sub_chars = extra_sub_chars + 1

! put in the duplicate single quote

              temp_sub_string(jj+extra_sub_chars:jj+extra_sub_chars) &
                = ''''

            END IF

          END DO

! Now TEMP_SUB_STRING has SUB_STRINGS( I/2 ) with dup. single quotes
! and EXTRA_SUB_CHARS is how many single quotes there were


! Shift remainder of FORMAT_STRING, if necessary
! NOTE: last EXTRA_SUB_CHARS characters will be lost
          IF (extra_sub_chars>0) format_string(sub_pos(i)+1+ &
            extra_fmt_chars+extra_sub_chars:) = format_string(sub_pos(i)+ &
            1+extra_fmt_chars:)

! Substitute

          format_string(sub_pos(i-1)+extra_fmt_chars: &
            sub_pos(i)+extra_fmt_chars+extra_sub_chars) = temp_sub_string

! Keep track of total extra characters added so far

          extra_fmt_chars = extra_fmt_chars + extra_sub_chars

        END DO SUBSTITUTE

      END SUBROUTINE edit_format

!*********************************************************************

      SUBROUTINE edit_message_format
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Executable Statements ..

        CALL edit_format(message_format)

      END SUBROUTINE edit_message_format

!*********************************************************************

      SUBROUTINE get_character(mssg,chars,num_chars,which,qerr)
! Displays mssg to user (explaining character choices), gets response
! from user, evaluates first non-blank character, returns index of
! response character in chars(1:num_chars) as which.
! chars should all be lower case.  Response letter is translated to
! lower case before evaluation.
! If no good answer in max_bad tries,
! qerr = .TRUE., otherwise qerr = .FALSE.
! Note: if num_chars is not positive and at most the length of chars,
! undefined behavior may result.
! .. Scalar arguments
! .. Local Parameters ..
! .. Local variables ..
! .. Intrinsic functions .. ( not all )
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: line_len = 80, max_bad = 3
! ..
! .. Scalar Arguments ..
        INTEGER, INTENT (IN) :: num_chars
        INTEGER, INTENT (OUT) :: which
        LOGICAL, INTENT (OUT) :: qerr
        CHARACTER (*), INTENT (IN) :: chars, mssg
! ..
! .. Local Scalars ..
        INTEGER :: i, ios, j
        CHARACTER (1) :: answer
        CHARACTER (line_len) :: line
! ..
! .. Intrinsic Functions ..
        INTRINSIC INDEX
! ..
! .. Executable Statements ..

! .. Executable code ..
        qerr = .FALSE.

GET_ANSWER: DO i = 1, max_bad

          IF (i>1) WRITE (stdout,'(/5X,''Please try again ...'')')

          WRITE (stdout,mssg)

          WRITE (stdout,'(1X,''Please enter one of ['',A,'']:'')', &
            advance='NO') chars(1:num_chars)

          CALL prompt

          READ (stdin,'(A)',iostat=ios) line

          IF (ios/=0) THEN

            WRITE (stdout,'(/5X,''Could not read the response.'')')

            CYCLE GET_ANSWER

          END IF

          DO j = 1, line_len

            IF (line(j:j)/=' ') EXIT

          END DO

          IF (j>line_len) THEN

            WRITE (stdout,'(/5X''A blank line is not allowed.'')')

            CYCLE GET_ANSWER

          END IF

          answer = low_case(line(j:j))

          which = INDEX(chars(1:num_chars),answer)

          IF (which>0) RETURN

          WRITE (stdout,'(/5X,A,'' is not allowed.'')') answer

        END DO GET_ANSWER

        qerr = .TRUE.

        RETURN

      END SUBROUTINE get_character

!*********************************************************************

      SUBROUTINE get_yn(mssg,qyes,qerr)
! Uses mssg to elicit a Y/N response from the user with get_character.
! If get_character fails, qerr == .TRUE., otherwise qerr == .FALSE.
! The value for qgetyn is undefined if qerr = .TRUE.
! otherwise is .TRUE. if user enters Y
!              .FALSE. if user enters N
! .. Dummy Scalars ..
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Local Scalars ..
        INTEGER :: which
! ..
! .. Scalar Arguments ..
        LOGICAL, INTENT (OUT) :: qerr, qyes
        CHARACTER (*), INTENT (IN) :: mssg
! ..
! .. Executable Statements ..

! .. Executable code ..
        CALL get_character(mssg,'yn',2,which,qerr)

        IF ( .NOT. qerr) qyes = (which==1)

        RETURN

      END SUBROUTINE get_yn

!*********************************************************************

      SUBROUTINE hold
! displays hold message to module unit STDOUT,
! waits for user input from module unit STDIN,
! and returns - all this should take place on one line.
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Local Scalars ..
        INTEGER :: ios
        CHARACTER (1) :: dummy_string
! ..
! .. Executable Statements ..

        WRITE (stdout, &
          '(/1X,''Press the Return or Enter key to continue ...'')')

        READ (stdin,'(A)',iostat=ios) dummy_string

        RETURN

      END SUBROUTINE hold

!*********************************************************************

      FUNCTION igtfun()
!-----------------------------------------------------------------------
!     Name: Get Free Unit Number
!     Description:
!     Returns a FORTRAN unit number that is not in current use.
!     Returns an invalid unit number (generally less than zero)
!     if no unassigned unit numbers are available.
!     Modified by BWB 22 Oct to use 5 and 6 as STDIN and STDOUT
!-----------------------------------------------------------------------
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Parameters ..
        INTEGER, PARAMETER :: one = 1, stdin = 5, stdout = 6, unummn = 1, &
          unummx = 20
! ..
! .. Function Return Value ..
        INTEGER :: igtfun
! ..
! .. Local Scalars ..
        INTEGER, SAVE :: lastun
        LOGICAL :: qcycle, qexist, qopen
! ..
! .. Data Statements ..
        DATA lastun/unummn/
! ..
! .. Executable Statements ..

        igtfun = lastun
        qcycle = .FALSE.

10      CONTINUE
        igtfun = igtfun - one

        IF (igtfun==stdin .OR. igtfun==stdout) GO TO 10

        IF (igtfun<unummn) THEN
          IF (qcycle) THEN
            igtfun = -1
            RETURN
          ELSE
            qcycle = .TRUE.
            igtfun = unummx
          END IF
        END IF

        IF ( .NOT. igtfun==lastun) GO TO 20
!     *** all available unit numbers are in use
        igtfun = unummn - one
        GO TO 30

20      CONTINUE
        INQUIRE (unit=igtfun,exist=qexist,opened=qopen)

        IF (qexist .AND. .NOT. qopen) GO TO 30

        GO TO 10

30      CONTINUE
        lastun = igtfun

        RETURN

      END FUNCTION igtfun

!*********************************************************************

      FUNCTION low_case(chr)
!-----------------------------------------------------------------------
!     Name: translate one character to LOWer CASE
!     Description:
!     If CHR is an upper case letter 'a..z' then returns the lower
!     case letter corresponding.
!     If CHR is not an upper case letter, then returns CHR.
!     Arguments:
!     CHR -> Character to be translated to upper case.
!            CHARACTER*1 CHR
!-----------------------------------------------------------------------
! .. Function type ..
! .. Dummy arg ..
! .. Local parameters ..
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Local Scalars ..
        INTEGER :: ix
        LOGICAL :: qnotuc
! ..
! .. Intrinsic Functions ..
        INTRINSIC INDEX, LGT, LLT
! ..
! .. Function Return Value ..
        CHARACTER (1) :: low_case
! ..
! .. Parameters ..
        CHARACTER (26), PARAMETER :: locase = &
          'abcdefghijklmnopqrstuvwxyz'
        CHARACTER (26), PARAMETER :: upcase = &
          'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
! ..
! .. Scalar Arguments ..
        CHARACTER (1), INTENT (IN) :: chr
! ..
! .. Executable Statements ..

        qnotuc = (LLT(chr,'A') .OR. LGT(chr,'Z'))

        IF (qnotuc) THEN

          low_case = chr

          RETURN

        END IF

        ix = INDEX(upcase,chr)

        IF (ix<=0) THEN

          low_case = chr

        ELSE

          low_case = locase(ix:ix)

        END IF

        RETURN

      END FUNCTION low_case

!*********************************************************************

      SUBROUTINE low_case_string(str)
!-----------------------------------------------------------------------
!     Name: LOWer CASE a whole STRING
!     Description:
!     For each CHR in STR,
!     If CHR is an upper case letter 'a..z' then it is replaced by the
!     lower case letter corresponding.
!     If CHR is not an upper case letter, then nothing happens to CHR.
!     Arguments:
!     STR <-> String to be translated to upper case.
!             CHARACTER*(*) STR
!-----------------------------------------------------------------------
! .. Dummy arg ..
! .. Local parameters ..
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Local Scalars ..
        INTEGER :: ix, j
        LOGICAL :: qnotuc
        CHARACTER (1) :: chr
! ..
! .. Intrinsic Functions ..
        INTRINSIC INDEX, LEN_TRIM, LGT, LLT
! ..
! .. Parameters ..
        CHARACTER (26), PARAMETER :: locase = &
          'abcdefghijklmnopqrstuvwxyz'
        CHARACTER (26), PARAMETER :: upcase = &
          'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
! ..
! .. Scalar Arguments ..
        CHARACTER (*), INTENT (INOUT) :: str
! ..
! .. Executable Statements ..

        DO j = 1, LEN_TRIM(str)

          chr = str(j:j)

          qnotuc = (LLT(chr,'A') .OR. LGT(chr,'Z'))

          IF (qnotuc) CYCLE

          ix = INDEX(upcase,chr)

          IF (ix<=0) CYCLE

          str(j:j) = locase(ix:ix)

        END DO

        RETURN

      END SUBROUTINE low_case_string

!*********************************************************************

      SUBROUTINE print_format(format_string,force,unit,unit_only)
! DMS Added optional arguments unit and unit_only to allow printing
!     in a report file.
!     unit --> Fortran unit associated with a report file.
!     unit_only --> if present, flags that printing will be
!                   performed only to the report file.
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL, INTENT (IN) :: unit
        LOGICAL, OPTIONAL, INTENT (IN) :: force, unit_only
        CHARACTER (*), INTENT (INOUT) :: format_string
! ..
! .. Local Scalars ..
        INTEGER :: output_unit
        LOGICAL :: force_print, qerr, qyes, want_help
! ..
! .. Parameters ..
        CHARACTER (*), PARAMETER :: print_question = '(1X,"Want &
          &next help message? (y/n)"/  1X,"If you don''t know &
          &what it is you probably want it.")'
! ..
! .. Intrinsic Functions ..
        INTRINSIC PRESENT
! ..
! .. Executable Statements ..

! See whether or not to really print
        force_print = .FALSE.

        IF (PRESENT(force)) THEN
          force_print = force
        END IF

        IF (print_off .AND. ( .NOT. force_print)) RETURN

! Set output unit

        IF (PRESENT(unit)) THEN
          output_unit = unit
        ELSE
          output_unit = -1
        END IF

! Do substitutions if necessary

        IF (num_subs>0) THEN
          CALL edit_format(format_string)
        END IF

! See if this message is desired

        IF (( .NOT. always_print) .AND. (print_level==2)) THEN

          CALL get_yn(print_question,qyes,qerr)

          IF (qerr) THEN
            want_help = .TRUE.
          ELSE
            want_help = qyes
          END IF

        END IF

! Print message

        IF ((force_print .OR. always_print .OR. print_level==1) .OR. &
            (print_level==2 .AND. want_help)) THEN

          IF (clear_screen_before_print) CALL clear_screen

! Always display on standard output

          IF ( .NOT. PRESENT(unit_only)) THEN
            WRITE (stdout,format_string)
          END IF

! If session has a report file, write in that file also

          IF (output_unit>0) THEN
            WRITE (output_unit,format_string)
          END IF

          format_printed = .TRUE.

        ELSE
          format_printed = .FALSE.
        END IF

      END SUBROUTINE print_format

!*********************************************************************

      SUBROUTINE print_message_format(force,unit,unit_only)
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Scalar Arguments ..
        INTEGER, OPTIONAL :: unit
        LOGICAL, OPTIONAL :: force, unit_only
! ..
! .. Executable Statements ..

        CALL print_format(message_format,force,unit,unit_only)

      END SUBROUTINE print_message_format

!*********************************************************************

      SUBROUTINE prompt
! Writes a prompt to module unit STDOUT, without advancing the line.
! .. Implicit None Statement ..
        IMPLICIT NONE
! ..
! .. Executable Statements ..

        WRITE (stdout,'(1X,''> '')',advance='NO')

        RETURN

      END SUBROUTINE prompt

!*********************************************************************

    END MODULE print_it
